package com.flyfish.framework.config;

import com.flyfish.framework.domain.AdminUserDetails;
import com.flyfish.framework.domain.po.User;
import com.flyfish.framework.enums.UserStatus;
import com.flyfish.framework.service.AuthenticationAuditor;
import com.flyfish.framework.service.AuthenticationLogger;
import com.flyfish.framework.service.ReactiveUserService;
import com.flyfish.framework.transform.ResultDataTransformer;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.web.server.WebFilterExchange;
import reactor.core.publisher.Mono;

import javax.annotation.Resource;
import java.util.Date;

/**
 * 提供对象写入，登录登出审计
 *
 * @author wangyu
 */
public class AuthenticationAuditorImpl extends ResultDataTransformer implements AuthenticationAuditor {

    @Resource
    private ReactiveUserService reactiveUserService;
    @Resource
    private AuthenticationLogger authenticationLogger;

    /**
     * 检查用户异常，无异常，清空错误次数，设置上次登录
     *
     * @param user 用户
     */
    @Override
    public Mono<Void> success(UserDetails user) {
        AdminUserDetails details = (AdminUserDetails) user;
        User updating = new User();
        updating.setId(details.getId());
        updating.setErrorCount(0);
        updating.setLastTime(new Date());
        return reactiveUserService.updateSelectiveById(updating)
                .then(Mono.fromRunnable(() -> authenticationLogger.success(user)));
    }

    /**
     * 登录失败的回调
     * 当输入错误的账号密码，记录错误次数，否则只记录日志
     *
     * @param webFilterExchange 请求信息
     * @param exception         异常
     */
    @Override
    public Mono<Void> error(WebFilterExchange webFilterExchange, AuthenticationException exception) {
        return webFilterExchange.getExchange()
                .getFormData()
                .flatMap(data -> {
                    // 当且仅当为凭据错误，尝试
                    if (exception instanceof BadCredentialsException) {
                        return reactiveUserService.findByUsername(data.getFirst("username"))
                                .flatMap(user -> {
                                    User updating = new User();
                                    updating.setId(user.getId());
                                    updating.setErrorCount(user.getErrorCount() + 1);
                                    if (updating.getErrorCount() >= 5) {
                                        updating.setStatus(UserStatus.LOCKED);
                                    }
                                    return reactiveUserService.updateSelectiveById(updating);
                                })
                                .map(user -> data);
                    }
                    return Mono.just(data);
                })
                .flatMap(data -> Mono.fromRunnable(() -> authenticationLogger.failure(data, exception)));
    }

    @Override
    public Mono<Void> logout(UserDetails userDetails) {
        return Mono.fromRunnable(() -> authenticationLogger.logout(userDetails));
    }
}
